"""
Calculations module APIs viewsets
"""

import json

from drf_yasg.utils import swagger_auto_schema
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.views import APIView

from djangophysics.converters.models import ConverterLoadError
from djangophysics.units.models import UnitSystem
from .exceptions import ExpressionCalculatorInitError
from .models import ExpressionCalculator
from .renderers import OperandUnitsRenderer
from .serializers import ExpressionSerializer, \
    CalculationPayloadSerializer, CalculationResultSerializer


class ValidateViewSet(APIView):
    """
    POST API View to validate formulas
    """

    @swagger_auto_schema(request_body=ExpressionSerializer)
    @action(['POST'], detail=False, url_path='', url_name="validate")
    def post(self, request, unit_system, *args, **kwargs):
        """
        Validate a formula with parameters
        :param request: HTTP request
        :param unit_system: Unit system to use for validation
        """
        if request.user and request.user.is_authenticated:
            us = UnitSystem(
                unit_system,
                user=request.user,
                key=request.data.get('key', None))
        else:
            us = UnitSystem(unit_system)
        exp = ExpressionSerializer(data=request.data)
        if exp.is_valid(unit_system=us, dimensions_only=True):
            return Response("Valid expression")
        else:
            return Response(json.dumps(exp.errors),
                            status=status.HTTP_406_NOT_ACCEPTABLE,
                            content_type="application/json")


class CalculationView(APIView):
    """
    Calculate a batch of formulas
    """
    renderer_classes = [OperandUnitsRenderer]

    @swagger_auto_schema(request_body=CalculationPayloadSerializer,
                         responses={200: CalculationResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency
        and date to a reference currency
        :param request: HTTP request
        """
        cps = CalculationPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST,
                            content_type="application/json")
        cp = cps.create(cps.validated_data)
        user = None
        if request.user and request.user.is_authenticated:
            user = request.user
        key = request.POST.get('key', None)
        try:
            calculator = ExpressionCalculator.load(
                user=user,
                key=key,
                id=cp.batch_id)
        except ConverterLoadError:
            calculator = ExpressionCalculator(
                id=cp.batch_id,
                unit_system=cp.unit_system,
                user=user,
                key=key
            )
        except ExpressionCalculatorInitError:
            return Response("Error initializing calculator",
                            status=status.HTTP_400_BAD_REQUEST)
        if cp.data:
            errors = calculator.add_data(data=cp.data)
            if errors:
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = calculator.convert()
            serializer = CalculationResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': calculator.id, 'status': calculator.status},
                            content_type="application/json")



"""
Calculations models
"""

import re
from datetime import date

import pint.systems
import uncertainties
from django.contrib.auth.models import User
from sympy import sympify, SympifyError

from djangophysics.units.models import UnitSystem, Dimension
from djangophysics.converters.models import BaseConverter, ConverterLoadError
from djangophysics.units.exceptions import DimensionNotFound, UnitSystemNotFound
from .exceptions import ExpressionCalculatorInitError


class Operand:
    """
    Operand in a formula
    """
    name = None
    value = None
    unit = None
    uncertainty = None

    def __init__(self,
                 name: str = None,
                 value: float = 0,
                 unit: str = None,
                 uncertainty: str = None):
        """
        Initialize Operand
        :param name: name of the operand that appears in the formula
        :param value: float magnitude
        :param unit: units
        """
        self.name = name
        self.value = value
        self.unit = unit
        self.uncertainty = uncertainty
        if self.validate():
            self.uvalue = uncertainties.ufloat(value, self.parse_uncertainty())
        else:
            self.uvalue = uncertainties.ufloat(0, 0)

    def validate(self):
        """
        Validate operand
        """
        if not self.name or self.value is None or self.unit is None:
            return False
        try:
            float(self.value)
        except ValueError:
            return False
        try:
            self.parse_uncertainty()
        except (ValueError, AttributeError):
            return False
        return True

    def get_unit(self, unit_system: UnitSystem) -> str:
        """
        Transform unit if unit is a dimension
        """
        # Find dimensions
        exp = r'(?P<dim>\[\w+\])'
        replace_dict = {}
        for dimension_name in re.findall(exp, self.unit):
            try:
                dim = Dimension(unit_system=unit_system, code=dimension_name)
                replace_dict[dimension_name] = str(dim.units().pop())
            except DimensionNotFound:
                pass
        # Replace in formula
        for key, item in replace_dict.items():
            self.unit = self.unit.replace(key, item)
        return self.unit

    def get_magnitude(self):
        return self.uvalue.n

    def get_uncertainty(self):
        return self.uvalue.s

    def parse_uncertainty(self):
        if not self.uncertainty:
            return 0
        if self.uncertainty.endswith('%'):
            return self.value * float(self.uncertainty[:-1]) * 0.01
        else:
            return float(self.uncertainty)


class ComputationError(Exception):
    pass


class Expression:
    """
    Expression with operands
    """
    expression = None
    operands = None
    out_units = None

    def __init__(
            self,
            expression: str,
            operands: [Operand],
            out_units: str = None):
        """
        Initialize Expression
        :param expression: string expression with placeholders
        :param operands: List of Operand corresponding to placeholders
        """
        self.expression = expression
        self.operands = operands
        self.out_units = out_units

    def _validate_syntax(self, unit_system: UnitSystem) -> (bool, str):
        """
        Syntactic validation of the expression
        :param unit_system: UnitSystem
        """
        if not self.expression:
            return False, "missing expression"
        try:
            sympify(
                self.expression.format(**{
                    v.name: v.value for v in self.operands}))
        except SympifyError:
            return False, "Improper expression"
        return True, ""

    def _validate_dimension(self, unit_system: UnitSystem) -> (bool, str):
        q_ = unit_system.ureg.Quantity
        kwargs = {v.name: q_(v.value, v.unit) for v in self.operands}
        try:
            result = unit_system.ureg.parse_expression(
                self.expression, **kwargs
            )
        except KeyError:
            return False, "Missing operand"
        except pint.errors.DimensionalityError:
            return False, "Incoherent dimensions"
        if self.out_units:
            try:
                result.to(self.out_units)
            except pint.errors.DimensionalityError:
                return False, "Incoherent output dimensions"
        return True, ''

    def validate(self, unit_system: UnitSystem) -> (bool, str):
        """
        Validate syntax and homogeneity
        :param unit_system: UnitSystem
        """
        valid_expression, error = self._validate_syntax(
            unit_system=unit_system
        )
        if not valid_expression:
            return False, error
        for var in self.operands:
            if not var.validate():
                return False, "invalid operand"
        valid_dimension, error = self._validate_dimension(
            unit_system=unit_system
        )
        if not valid_dimension:
            return False, error
        return True, ''

    def evaluate(self, unit_system: UnitSystem) -> pint.Quantity:
        """
        Validate formula
        """
        q_ = unit_system.ureg.Quantity
        is_valid, error = self.validate(unit_system=unit_system)
        if is_valid:
            kwargs = {v.name: q_(v.uvalue, v.unit) for v in self.operands}
            result = unit_system.ureg.parse_expression(
                self.expression, **kwargs
            )
            if self.out_units:
                return result.to(self.out_units)
            else:
                return result
        else:
            raise ComputationError(f"Invalid formula: {error}")


class CalculationPayload:
    """
    Calculation payload: caculate expressions
    """
    data = None
    unit_system = ''
    user = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(
            self,
            unit_system: UnitSystem,
            key: str = None,
            data: [] = None,
            batch_id: str = None,
            eob: bool = False):
        """
        Initialize payload
        """
        self.data = data
        self.unit_system = unit_system
        self.key = key
        self.batch_id = batch_id
        self.eob = eob


class CalculationResultDetail:
    """
    Details of an evaluation
    """
    expression = None
    operands = None
    magnitude = None
    uncertainty = None
    unit = None

    def __init__(self, expression: str, operands: [],
                 magnitude: uncertainties.ufloat, unit: str):
        """
        Initialize detail
        :param expression: Expression in the form of a string
        e.g.: 3*{a} + 2 * {b}
        :param operands: List of Operands
        :param magnitude: result of the calculation
        :param units: dimension of the result
        :param uncertainty: uncertainty of the calculation
        """
        self.expression = expression
        self.operands = operands
        self.magnitude = magnitude.n
        self.uncertainty = magnitude.s
        self.unit = unit


class CalculationResultError:
    """
    Error details of a wrong calculation
    """
    expression = None
    operands = None
    calc_date = None
    error = None

    def __init__(
            self,
            expression: str,
            operands: [],
            calc_date: date,
            error: str):
        """
        Initialize error detail
        :param expression: Expression in the form of a string
        :param operands: List of Operands
        :param date: date of the evaluation
        :param error: Error description
        """
        self.expression = expression
        self.operands = operands
        self.calc_date = calc_date
        self.error = error


class CalculationResult:
    """
    Result of a batch of evaluations
    """
    id = None
    detail = []
    status = None
    errors = []

    def __init__(self,
                 id: str = None,
                 detail: [CalculationResultDetail] = None,
                 status: str = BaseConverter.INITIATED_STATUS,
                 errors: [CalculationResultError] = None):
        """
        Initialize result
        :param id: ID of the batch
        :param detail: List of CalculationResultDetail
        :param status: status of the batch
        :param errors: List of CalculationResultErrors
        """
        self.id = id
        self.detail = detail or []
        self.status = status
        self.errors = errors or []

    def end_batch(self):
        """
        Puts a finall status on the batch
        """
        if self.errors:
            self.status = BaseConverter.WITH_ERRORS
        else:
            self.status = BaseConverter.FINISHED
        return self.status


class ExpressionCalculator(BaseConverter):
    """
    Conversion between units
    """
    unit_system = None

    def __init__(
            self,
            unit_system: str,
            user: User = None,
            key: str = '',
            id: str = None):
        """
        Initiate ExpressionCalculator
        :param unit_system: unit system name
        :param user: User
        :param key: key of user
        :param id: ID of the batch
        """
        try:
            super().__init__(id=id)
            self.unit_system = unit_system
            self.user = user
            self.key = key
            self.system = UnitSystem(
                system_name=unit_system,
                user=self.user,
                key=self.key)
        except UnitSystemNotFound as e:
            raise ExpressionCalculatorInitError from e

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super().add_data(data)
        return errors

    def check_data(self, data):
        """
        Validates that the data contains
        - system (str)
        - unit (str)
        - value (float)
        - date (YYYY-MM-DD)
        """
        from .serializers import ExpressionSerializer
        errors = []
        for line in data:
            serializer = ExpressionSerializer(data=line)
            if serializer.is_valid(unit_system=self.system):
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    @classmethod
    def load(
            cls,
            id: str,
            user: User = None,
            key: str = None) -> BaseConverter:
        """
        Load converter from batch
        :param id: ID of the batch
        :param user: Users object
        :param key: User defined category
        """
        try:
            uc = super().load(id)
            uc.system = UnitSystem(
                system_name=uc.unit_system,
                user=user,
                key=key)
            return uc
        except (UnitSystemNotFound, KeyError) as e:
            raise ConverterLoadError from e

    def save(self):
        """
        Save converter to caching system
        """
        system = self.system
        self.system = None
        super(ExpressionCalculator, self).save()
        self.system = system

    def convert(self) -> CalculationResult:
        """
        Converts data to base unit in base system
        """
        result = CalculationResult(id=self.id)
        for expression in self.data:
            valid, exp_error = expression.validate(unit_system=self.system)
            if not valid:
                error = CalculationResultError(
                    expression=expression.expression,
                    operands=expression.operands,
                    calc_date=date.today(),
                    error=exp_error
                )
                result.errors.append(error)
                continue
            try:
                out = expression.evaluate(unit_system=self.system)
            except ComputationError as e:
                error = CalculationResultError(
                    expression=expression.expression,
                    operands=expression.operands,
                    calc_date=date.today(),
                    error=str(e)
                )
                result.errors.append(error)
                continue
            detail = CalculationResultDetail(
                expression=expression.expression,
                operands=expression.operands,
                magnitude=out.magnitude,
                unit=out.units
            )
            result.detail.append(detail)
        self.end_batch(result.end_batch())
        return result



"""
Units module APIs viewsets
"""

import logging

from django.db import models
from django.http import HttpResponseForbidden, HttpRequest
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from django_filters import rest_framework as filters
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import permissions, status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_400_BAD_REQUEST
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet, ModelViewSet

from djangophysics.converters.models import ConverterLoadError
from djangophysics.converters.serializers import ConverterResultSerializer
from djangophysics.core.helpers import validate_language
from djangophysics.core.pagination import PageNumberPagination
from . import DIMENSIONS
from .exceptions import UnitConverterInitError, UnitSystemNotFound, \
    UnitNotFound, DimensionNotFound, UnitValueError
from .filters import CustomUnitFilter
from .forms import CustomUnitForm
from .models import UnitSystem, UnitConverter, Dimension, CustomUnit
from .permissions import CustomUnitObjectPermission
from .serializers import UnitSerializer, UnitSystemSerializer, \
    UnitConversionPayloadSerializer, DimensionSerializer, \
    DimensionWithUnitsSerializer, CustomUnitSerializer


class UnitSystemViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'system_name'

    language_header = openapi.Parameter(
        'Accept-Language', openapi.IN_HEADER,
        description="language",
        type=openapi.TYPE_STRING)
    language = openapi.Parameter(
        'language', openapi.IN_QUERY,
        description="language",
        type=openapi.TYPE_STRING)

    unit_systems_response = openapi.Response(
        'List of unit systems',
        UnitSystemSerializer)
    unit_system_response = openapi.Response(
        'Dimensions and units in a system',
        UnitSystemSerializer)

    ordering = openapi.Parameter(
        'ordering', openapi.IN_QUERY,
        description="Sort on name Prefix with - for descending sort",
        type=openapi.TYPE_STRING)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(
        manual_parameters=[language, language_header, ordering],
        responses={200: unit_systems_response})
    def list(self, request):
        """
        List UnitSystems
        """
        ordering = request.GET.get('ordering', 'system_name')
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['system_name']:
            ordering = 'system_name'
        language = validate_language(request.GET.get('language',
                                                     request.LANGUAGE_CODE))
        us = UnitSystem(fmt_locale=language)
        us = [{'system_name': s} for s in sorted(us.available_systems(),
                                                 reverse=descending)]
        return Response(us, content_type="application/json")

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header],
                         responses={200: unit_system_response})
    def retrieve(self, request, system_name):
        """
        Retrieve UnitSystem
        """
        language = validate_language(request.GET.get('language',
                                                     request.LANGUAGE_CODE))
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            serializer = UnitSystemSerializer(us, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except UnitSystemNotFound as e:
            return Response("Unknown unit system: " + str(e),
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(
        manual_parameters=[language, language_header, ordering],
        responses={200: DimensionSerializer})
    @action(methods=['GET'], detail=True,
            name='dimensions', url_path='dimensions')
    def dimensions(self, request, system_name):
        """
        List Dimensions of a UnitSystem
        """
        language = validate_language(request.GET.get('language',
                                                     request.LANGUAGE_CODE))
        ordering = request.GET.get('ordering', 'name')
        try:
            us = UnitSystem(system_name=system_name, fmt_locale=language)
            serializer = DimensionSerializer(us.available_dimensions(
                ordering=ordering),
                many=True,
                context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except UnitSystemNotFound as e:
            return Response("Unknown unit system: " + str(e),
                            status=HTTP_404_NOT_FOUND)


class UnitViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'unit_name'
    language_header = openapi.Parameter(
        'Accept-Language', openapi.IN_HEADER,
        description="language",
        type=openapi.TYPE_STRING)
    language = openapi.Parameter(
        'language', openapi.IN_QUERY,
        description="language",
        type=openapi.TYPE_STRING)
    dimension = openapi.Parameter(
        'dimension', openapi.IN_QUERY,
        description="Unit dimension",
        type=openapi.TYPE_STRING)
    key = openapi.Parameter(
        'key',
        openapi.IN_QUERY,
        description="key",
        type=openapi.TYPE_STRING)
    units_response = openapi.Response(
        'List of units in a system',
        UnitSerializer)
    unit_response = openapi.Response(
        'Detail of a unit',
        UnitSerializer)
    dimension_response = openapi.Response(
        'List of units per dimension',
        DimensionWithUnitsSerializer)
    ordering = openapi.Parameter(
        'ordering', openapi.IN_QUERY,
        description="Sort on fields name or code. "
                    "Prefix with - for descending sort",
        type=openapi.TYPE_STRING)

    @swagger_auto_schema(manual_parameters=[dimension, key,
                                            ordering,
                                            language, language_header],
                         responses={200: units_response})
    def list(self, request: HttpRequest, system_name: str):
        """
        List Units, ordered, filtered by key or dimension,
        translated in language
        """
        language = validate_language(request.GET.get(
            'language', request.LANGUAGE_CODE))
        ordering = request.GET.get('ordering', 'name')
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['code', 'name']:
            ordering = 'name'
        try:
            key = request.GET.get('key', None)
            user = request.user if \
                hasattr(request, 'user') and \
                request.user.is_authenticated else None
            us = UnitSystem(
                system_name=system_name,
                fmt_locale=language,
                user=user,
                key=key)
            units = []
            dimension_param = request.GET.get(key='dimension')
            if dimension_param:
                try:
                    dimension = Dimension(unit_system=us,
                                          code=dimension_param)
                    units = dimension.units(user=user, key=key)
                except DimensionNotFound:
                    return Response('Invalid dimension filter',
                                    status=status.HTTP_400_BAD_REQUEST)
            else:
                available_units = us.available_unit_names()
                if available_units:
                    units = [us.unit(unit_name=unit_name)
                             for unit_name in available_units]
            units = sorted(units, key=lambda x: getattr(x, ordering),
                           reverse=descending)
            serializer = UnitSerializer(
                units,
                many=True,
                context={'request': request})
            return Response(serializer.data)
        except UnitSystemNotFound:
            return Response('Invalid Unit System',
                            status=status.HTTP_404_NOT_FOUND)

    @swagger_auto_schema(manual_parameters=[key, language, language_header],
                         responses={200: dimension_response})
    @action(['GET'], detail=False,
            name='units per dimension', url_path='per_dimension')
    def list_per_dimension(self, request: HttpRequest, system_name: str):
        """
        List Units grouped by dimension, filter on key, translated
        """
        language = validate_language(request.GET.get('language',
                                                     request.LANGUAGE_CODE))
        try:
            key = request.GET.get('key', None)
            user = request.user if \
                hasattr(request, 'user') and \
                request.user.is_authenticated else None
            us = UnitSystem(
                system_name=system_name,
                fmt_locale=language,
                user=user,
                key=key)
            dimensions = [Dimension(unit_system=us, code=code)
                          for code in DIMENSIONS.keys()]
            serializer = DimensionWithUnitsSerializer(
                dimensions,
                many=True,
                context={'request': request})
            return Response(serializer.data)
        except UnitSystemNotFound as e:
            logging.warning(str(e))
            return Response('Invalid Unit System',
                            status=status.HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[key, language, language_header],
                         responses={200: unit_response})
    def retrieve(self, request: HttpRequest, system_name: str, unit_name: str):
        """
        Get unit information for unit in unit system
        """
        language = validate_language(request.GET.get('language',
                                                     request.LANGUAGE_CODE))
        try:
            key = request.GET.get('key', None)
            user = request.user if \
                hasattr(request, 'user') and \
                request.user.is_authenticated else None
            us = UnitSystem(
                system_name=system_name,
                fmt_locale=language,
                user=user,
                key=key)
            unit = us.unit(unit_name=unit_name)
            serializer = UnitSerializer(unit, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (UnitSystemNotFound, UnitNotFound):
            return Response("Unknown unit", status=HTTP_404_NOT_FOUND)

    @swagger_auto_schema(
        manual_parameters=[language, language_header, ordering],
        responses={200: units_response})
    @action(methods=['GET'], detail=True,
            url_path='compatible', url_name='compatible_units')
    def compatible_units(self,
                         request: HttpRequest,
                         system_name: str,
                         unit_name: str):
        """
        List compatible Units
        """
        language = validate_language(request.GET.get('language',
                                                     request.LANGUAGE_CODE))
        ordering = request.GET.get('ordering', 'name')
        descending = False
        if ordering and ordering[0] == '-':
            ordering = ordering[1:]
            descending = True
        if ordering not in ['code', 'name']:
            ordering = 'name'
        try:
            key = request.GET.get('key', None)
            user = request.user if \
                hasattr(request, 'user') and \
                request.user.is_authenticated else None
            us = UnitSystem(
                system_name=system_name,
                fmt_locale=language,
                user=user,
                key=key)
            unit = us.unit(unit_name=unit_name)
            compatible_units = sorted([us.unit(unit_name=cunit) for cunit in
                                       map(str, unit.unit.compatible_units())],
                                      key=lambda x: getattr(x, ordering),
                                      reverse=descending)
            serializer = UnitSerializer(
                compatible_units,
                many=True,
                context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except (UnitSystemNotFound, UnitNotFound):
            return Response("Unknown unit", status=HTTP_404_NOT_FOUND)


class ConvertView(APIView):
    """
    Convert Units API
    """

    @swagger_auto_schema(request_body=UnitConversionPayloadSerializer,
                         responses={200: ConverterResultSerializer})
    @action(['POST'], detail=False, url_path='', url_name="convert")
    def post(self, request, *args, **kwargs):
        """
        Converts a list of amounts with currency
        and date to a reference currency
        :param request: HTTP request
        """
        cps = UnitConversionPayloadSerializer(data=request.data)
        if not cps.is_valid():
            return Response(cps.errors, status=HTTP_400_BAD_REQUEST,
                            content_type="application/json")
        cp = cps.create(cps.validated_data)
        user = None
        if request.user and request.user.is_authenticated:
            user = request.user
        key = request.POST.get('key', None)
        try:
            converter = UnitConverter.load(user=user, key=key, id=cp.batch_id)
        except ConverterLoadError:
            converter = UnitConverter(
                id=cp.batch_id,
                base_system=cp.base_system,
                base_unit=cp.base_unit,
                user=user,
                key=key
            )
        except UnitConverterInitError:
            return Response("Error initializing converter",
                            status=status.HTTP_400_BAD_REQUEST)
        if cp.data:
            errors = converter.add_data(data=cp.data)
            if errors:
                return Response(errors, status=HTTP_400_BAD_REQUEST)
        if cp.eob or not cp.batch_id:
            result = converter.convert()
            serializer = ConverterResultSerializer(result)
            return Response(serializer.data, content_type="application/json")
        else:
            return Response({'id': converter.id, 'status': converter.status},
                            content_type="application/json")


class CustomUnitViewSet(ModelViewSet):
    """
    Custom Units API
    """
    queryset = CustomUnit.objects.all()
    serializer_class = CustomUnitSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = CustomUnitFilter
    pagination_class = PageNumberPagination
    permission_classes = [CustomUnitObjectPermission,
                          permissions.IsAuthenticated]
    display_page_controls = True
    lookup_url_param = 'system_name'

    user = openapi.Parameter(
        'user',
        openapi.IN_QUERY,
        description="Filter on user rates",
        type=openapi.TYPE_BOOLEAN)
    key = openapi.Parameter(
        'key',
        openapi.IN_QUERY,
        description="Filter on user defined category",
        type=openapi.TYPE_STRING)
    unit_system = openapi.Parameter(
        'unit_system',
        openapi.IN_QUERY,
        description="Filter on unit system",
        type=openapi.TYPE_STRING)
    code = openapi.Parameter(
        'code',
        openapi.IN_QUERY,
        description="Filter on unit code",
        type=openapi.TYPE_STRING)
    name = openapi.Parameter(
        'name',
        openapi.IN_QUERY,
        description="Filter on unit name",
        type=openapi.TYPE_STRING)
    relation = openapi.Parameter(
        'relation',
        openapi.IN_QUERY,
        description="Filter on relation to base units",
        type=openapi.TYPE_STRING)
    symbol = openapi.Parameter(
        'symbol',
        openapi.IN_QUERY,
        description="Filter on unit symbol",
        type=openapi.TYPE_STRING)
    alias = openapi.Parameter(
        'alias',
        openapi.IN_QUERY,
        description="Filter on unit alias",
        type=openapi.TYPE_STRING)
    ordering = openapi.Parameter(
        'ordering',
        openapi.IN_QUERY,
        description="Sort on code, name, relation, symbol, alias. "
                    "Prefix with - for descending sort",
        type=openapi.TYPE_STRING)

    def get_queryset(self):
        """
        Filter units based on authenticated user
        """
        qs = super(CustomUnitViewSet, self).get_queryset()
        system_name = self.kwargs.get(self.lookup_url_param, 'SI')
        if self.request.user and self.request.user.is_authenticated:
            qs = qs.filter(
                models.Q(user=self.request.user) | models.Q(user__isnull=True)
            )
            if self.request.GET.get('key'):
                qs = qs.filter(key=self.request.GET.get('key'))
        else:
            qs = qs.filter(models.Q(user__isnull=True))
        qs = qs.filter(unit_system__iexact=system_name.lower())
        return qs

    @swagger_auto_schema(manual_parameters=[
        user, key, unit_system, code, name, relation, symbol, alias, ordering],
        responses={200: CustomUnitSerializer})
    def list(self, request, *args, **kwargs):
        """
        List rates
        """
        return super().list(request, *args, **kwargs)

    def create(self, request: HttpRequest, system_name: str, *args, **kwargs):
        """
        Create CustomUnit
        """
        cu_form = CustomUnitForm(request.data)
        if cu_form.is_valid():
            cu = cu_form.save(commit=False)
            try:
                UnitSystem(
                    system_name=system_name,
                    user=request.user,
                    key=cu.key)
            except UnitSystemNotFound:
                return Response("Invalid unit system",
                                status=status.HTTP_400_BAD_REQUEST)
            if request.user and request.user.is_authenticated:
                if CustomUnit.objects.filter(
                        code=cu.code,
                        user=request.user,
                        key=cu.key).exists():
                    return Response("Custom unit already exists",
                                    status=status.HTTP_409_CONFLICT)
                cu.user = request.user
                cu.unit_system = system_name
                try:
                    cu.save()
                except (UnitValueError, ValueError) as e:
                    return Response(str(e),
                                    status=status.HTTP_400_BAD_REQUEST)
                serializer = CustomUnitSerializer(cu)
                return Response(serializer.data,
                                status=status.HTTP_201_CREATED)
            else:
                return HttpResponseForbidden()
        else:
            return Response(cu_form.errors,
                            status=status.HTTP_400_BAD_REQUEST,
                            content_type="application/json")



"""
Calculations app configuration
"""

from django.apps import AppConfig


class CalculationsConfig(AppConfig):
    """
    Calculations configuration
    """
    name = "djangophysics.calculations"
    verbose_name = "Calculations"



"""
Units tests
"""
import json
import uuid

import pint
from django.core.cache import cache
from django.test import TestCase
from rest_framework import status
from rest_framework.test import APIClient

from djangophysics.units.models import UnitSystem

from .models import ExpressionCalculator, Operand
from .serializers import ExpressionSerializer


class ExpressionTest(TestCase):
    """
    Test Expression
    """

    def setUp(self):
        """
        Setup test environment
        """
        self.us = UnitSystem(system_name='SI')

    def tearDown(self):
        """
        tear down environment
        """
        self.us = None

    def test_valid_serializer_payload(self):
        """
        Test serializer validation
        """
        payload = {
            'expression': '(3*{a}+15*{b})*6*{c}',
            'operands': [
                {
                    'name': 'a',
                    'value': 0.1,
                    'unit': 'kg'
                },
                {
                    'name': 'b',
                    'value': 15,
                    'unit': 'g'
                },
                {
                    'name': 'c',
                    'value': 12,
                    'unit': 's'
                }
            ]
        }
        es = ExpressionSerializer(data=payload)
        self.assertTrue(es.is_valid(unit_system=self.us))

    def test_invalid_serializer_expression(self):
        """
        Test invalid payload
        """
        payload = {
            'expression': '(3*{a}+15*{b}) -',
            'operands': [
                {
                    'name': 'a',
                    'value': 0.1,
                    'unit': 'kg'
                },
                {
                    'name': 'b',
                    'value': 15,
                    'unit': 'g'
                },
                {
                    'name': 'c',
                    'value': 12,
                    'unit': 's'
                }
            ]
        }
        es = ExpressionSerializer(data=payload)
        self.assertFalse(es.is_valid(unit_system=self.us))

    def test_invalid_serializer_coherency(self):
        """
        Test Quantity coherency
        """
        payload = {
            'expression': '(3*{a}+15*{b})-c',
            'operands': [
                {
                    'name': 'a',
                    'value': 0.1,
                    'unit': 'kg'
                },
                {
                    'name': 'b',
                    'value': 15,
                    'unit': 'g'
                },
                {
                    'name': 'c',
                    'value': 12,
                    'unit': 's'
                }
            ]
        }
        es = ExpressionSerializer(data=payload)
        self.assertFalse(es.is_valid(unit_system=self.us))

    def test_invalid_serializer_parameters(self):
        """
        Test wrong parameters
        """
        payload = {
            'expression': '(3*{a}+15*{b})-c',
            'operands': [
                {
                    'name': 'a',
                    'value': 0.1,
                    'unit': 'kg'
                },
                {
                    'name': 'b',
                    'value': 15,
                    'unit': 'g'
                }
            ]
        }
        es = ExpressionSerializer(data=payload)
        self.assertFalse(es.is_valid(unit_system=self.us))

    def test_empty_serializer_expression(self):
        """
        Test empty expression in payload
        """
        payload = {
            'expression': '',
            'operands': [
                {
                    'name': 'a',
                    'value': 0.1,
                    'unit': 'kg'
                },
                {
                    'name': 'b',
                    'value': 15,
                    'unit': 'g'
                }
            ]
        }
        es = ExpressionSerializer(data=payload)
        self.assertFalse(es.is_valid(unit_system=self.us))

    def test_empty_serializer_operands(self):
        """
        Test empty operands list in payload
        """
        payload = {
            'expression': '(3*{a}+15*{b})-c',
            'operands': []
        }
        es = ExpressionSerializer(data=payload)
        self.assertFalse(es.is_valid(unit_system=self.us))

    def test_formula_validation_request(self):
        """
        Test formula syntax validation
        """
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/validate/',
            data=json.dumps({
                'expression': "3*{a}+15*{b}",
                "operands": [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 15,
                        "unit": "g"
                    }
                ]
            }),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)


class OperandTest(TestCase):
    """
    Test Operand
    """

    def setUp(self):
        """
        Setup unit system
        """
        self.us = UnitSystem(system_name='SI')

    def test_validate(self):
        """
        Test operand validation
        """
        op = Operand(
            name='test',
            value=0,
            unit='m/s'
        )
        self.assertTrue(op.validate())
        op = Operand(
            name='',
            value=1,
            unit='m/s'
        )
        self.assertFalse(op.validate())
        op = Operand(
            name='test',
            value=None,
            unit='m/s'
        )
        self.assertFalse(op.validate())
        op = Operand(
            name='test',
            value='a',
            unit='m/s'
        )
        self.assertFalse(op.validate())

    def test_get_unit_units(self):
        """
        Test units of operand
        """
        op = Operand(
            name='toto',
            value=15,
            unit='m/s'
        )
        self.assertTrue(op.validate())
        self.assertEqual(op.get_unit(self.us), 'm/s')

    def test_get_unit_uncertainty_and_magnitude(self):
        """
        Test uncertainty of operand
        """
        op = Operand(
            name='toto',
            value=15,
            unit='m/s',
            uncertainty="0.01"
        )
        self.assertTrue(op.validate())
        self.assertEqual(op.get_uncertainty(), 0.01)
        self.assertEqual(op.get_magnitude(), 15)

    def test_get_unit_uncertainty_percentage_and_magnitude(self):
        """
        Test uncertainty with a percentage
        """
        op = Operand(
            name='toto',
            value=15,
            unit='m/s',
            uncertainty="10%"
        )
        self.assertTrue(op.validate())
        self.assertEqual(op.get_uncertainty(), 15 * 0.1)
        self.assertEqual(op.get_magnitude(), 15)

    def test_get_unit_dimensions(self):
        """
        Test dimensions of units
        """
        q_ = self.us.ureg.Quantity
        op = Operand(
            name='toto',
            value=15,
            unit='[mass]/[time]'
        )
        self.assertIsInstance(q_(1, op.get_unit(self.us)), q_)

    def test_get_unit_mixed(self):
        """
        Test operand with dimensions
        """
        q_ = self.us.ureg.Quantity
        op = Operand(
            name='toto',
            value=15,
            unit='[mass]/[time]*L'
        )
        self.assertIsInstance(q_(1, op.get_unit(self.us)), q_)

    def test_get_unit_mixed_custom(self):
        """
        Test Operand with custom units
        """
        q_ = self.us.ureg.Quantity
        op = Operand(
            name='toto',
            value=15,
            unit='[mass]/[time]*L*plouf'
        )
        self.assertRaises(
            pint.errors.UndefinedUnitError,
            q_,
            1,
            op.get_unit(self.us)
        )


class ExpressionAPITest(TestCase):
    """
    Test Expression API
    """

    def setUp(self):
        """
        Setup test environment
        """
        self.us = UnitSystem(system_name='SI')

    def test_formula_validation_variable_exception_request(self):
        """
        Test formaula validation API
        """
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/validate/',
            data=json.dumps({
                'expression': "3*{a}+15*{b}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                ]
            }),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_406_NOT_ACCEPTABLE)

    def test_formula_validation_expression_exception_request(self):
        """
        Test wrong formula validation API
        """
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/validate/',
            data=json.dumps({
                'expression': "3*{a}+15*{b}+",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 15,
                        "unit": "g"
                    }
                ]
            }),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_406_NOT_ACCEPTABLE)


class ExpressionCalculatorTest(TestCase):
    """
    Test expression calculator
    """
    base_system = 'SI'
    base_unit = 'meter'

    def setUp(self) -> None:
        """
        Setup environment
        """
        self.calculator = ExpressionCalculator(unit_system='SI')
        self.unit_system = UnitSystem(system_name='SI')
        self.expressions = [
            {
                'expression': "3*{a}+15*{b}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 15,
                        "unit": "g"
                    }
                ]
            },
            {
                'expression': "3*{a}+15*{b}+1000*{c}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 150,
                        "unit": "g"
                    },
                    {
                        "name": "c",
                        "value": 250,
                        "unit": "mg"
                    },
                ]
            }
        ]
        self.expression_to_unit = [
            {
                'expression': "3*{a}+15*{b}",
                'operands': [
                    {
                        "name": "a",
                        "value": 200,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 150000,
                        "unit": "g"
                    }
                ],
                'out_units': 'pound'
            },
            {
                'expression': "3*{a}+15*{b}+1000*{c}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 75,
                        "unit": "g"
                    },
                    {
                        "name": "c",
                        "value": 125,
                        "unit": "mg"
                    },
                ],
                'out_units': 'milligram'
            }
        ]
        self.trash_expressions = [
            {
                'expression': "3*{a}+15*{b}+",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 15,
                        "unit": "g"
                    }
                ]
            },
            {
                'expression': "3*{a}+15*{b}+1000*{c}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 150,
                        "unit": "g"
                    },
                    {
                        "name": "c",
                        "value": 250,
                        "unit": "s"
                    },
                ]
            },
            {
                'expression': "3*{a}+15*{b}+1000*{c}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 75,
                        "unit": "g"
                    },
                    {
                        "name": "c",
                        "value": 125,
                        "unit": "mg"
                    },
                ],
                'out_units': 'L'
            }
        ]

    def tearDown(self):
        """
        tear down environment
        """
        self.unit_system = None

    def test_created(self):
        """
        Test calculator creation
        """
        self.assertEqual(
            self.calculator.status,
            self.calculator.INITIATED_STATUS)

    def test_add_data(self):
        """
        Test adding data to calculator
        """
        errors = self.calculator.add_data(self.expressions)
        self.assertEqual(errors, [])
        self.assertEqual(
            self.calculator.status,
            self.calculator.INSERTING_STATUS)
        self.assertIsNotNone(cache.get(self.calculator.id))

    def test_trash_quantities(self):
        """
        Test adding trash to the calculator
        """
        calculator = ExpressionCalculator(unit_system='SI')
        errors = calculator.add_data(self.trash_expressions)
        self.assertEqual(len(errors), 3)
        self.assertEqual([list(c.keys())[0] for c in errors],
                         ['expression', 'expression', 'out_units'])

    def test_add_empty_data(self):
        """
        Test adding empty data to the calculator
        """
        calculator = ExpressionCalculator(unit_system='SI')
        errors = calculator.add_data(data=None)
        self.assertEqual(len(errors), 1)

    def test_computation(self):
        """
        Test computation
        """
        self.calculator.add_data(self.expressions)
        result = self.calculator.convert()
        self.assertEqual(result.id, self.calculator.id)
        self.assertEqual(self.calculator.status, self.calculator.FINISHED)
        self.assertEqual(len(result.errors), 0)
        self.assertEqual(len(result.detail), len(self.calculator.data))

    def test_computation_and_conversion(self):
        """
        Test computation
        """
        self.calculator.add_data(self.expression_to_unit)
        result = self.calculator.convert()
        self.assertEqual(result.id, self.calculator.id)
        self.assertEqual(self.calculator.status, self.calculator.FINISHED)
        self.assertEqual(len(result.errors), 0)
        self.assertEqual(len(result.detail), len(self.calculator.data))
        self.assertEqual(result.detail[0].unit, 'pound')
        self.assertEqual(result.detail[1].unit, 'milligram')


class ExpressionCalculatorAPITest(TestCase):
    """
    Test expression calculator API
    """
    base_system = 'SI'
    base_unit = 'meter'

    def setUp(self) -> None:
        """
        Setup environment
        """
        self.calculator = ExpressionCalculator(unit_system='SI')
        self.unit_system = UnitSystem(system_name='SI')
        self.expressions = [
            {
                'expression': "3*{a}+15*{b}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 15,
                        "unit": "g"
                    }
                ]
            },
            {
                'expression': "3*{a}+15*{b}+1000*{c}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 150,
                        "unit": "g"
                    },
                    {
                        "name": "c",
                        "value": 250,
                        "unit": "mg"
                    },
                ]
            }
        ]
        self.expression_to_unit = [
            {
                'expression': "3*{a}+15*{b}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 15,
                        "unit": "g"
                    }
                ],
                'out_units': 'ton'
            },
            {
                'expression': "3*{a}+15*{b}+1000*{c}",
                'operands': [
                    {
                        "name": "a",
                        "value": 0.1,
                        "unit": "kg"
                    },
                    {
                        "name": "b",
                        "value": 150,
                        "unit": "g"
                    },
                    {
                        "name": "c",
                        "value": 250,
                        "unit": "mg"
                    },
                ],
                'out_units': 'hundredweight'
            }
        ]

    def test_convert_request(self):
        """
        Test calculate API
        """
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/calculate/',
            data={
                'data': self.expressions,
                'unit_system': 'SI',
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            len(response.json().get('detail')),
            len(self.expressions))
        self.assertEqual(response.json().get('detail')[0]['magnitude'],
                         0.525)

    def test_convert_to_unit_request(self):
        """
        Test calculate API
        """
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/calculate/',
            data={
                'data': self.expression_to_unit,
                'unit_system': 'SI',
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json().get('detail')),
                         len(self.expression_to_unit))
        self.assertEqual(response.json().get('detail')[0]['unit'],
                         'ton')
        self.assertEqual(response.json().get('detail')[1]['unit'],
                         'hundredweight')

    def test_convert_batch_request(self):
        """
        Test batch computation
        """
        batch_id = uuid.uuid4()
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/calculate/',
            data={
                'data': self.expressions,
                'batch_id': batch_id,
                'unit_system': 'SI',

            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('id', response.json())
        self.assertEqual(response.json().get('status'),
                         ExpressionCalculator.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))
        response = client.post(
            '/units/SI/formulas/calculate/',
            data={
                'data': self.expressions,
                'unit_system': 'SI',
                'batch_id': batch_id,
                'eob': True
            },
            format='json')
        self.assertEqual(response.json().get('status'),
                         ExpressionCalculator.FINISHED)
        self.assertEqual(len(response.json().get('detail')),
                         2 * len(self.expressions))

    def test_watch_request(self):
        """
        Test observation of the calculator
        """
        batch_id = uuid.uuid4()
        client = APIClient()
        response = client.post(
            '/units/SI/formulas/calculate/',
            data={
                'data': self.expressions,
                'batch_id': batch_id,
                'unit_system': 'SI',
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('id', response.json())
        self.assertEqual(response.json().get('status'),
                         ExpressionCalculator.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))
        response = client.get(f'/watch/{str(batch_id)}/', format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json().get('status'),
                         ExpressionCalculator.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))



"""
Calculations module URLs
"""

from django.urls import path

from .viewsets import ValidateViewSet, CalculationView

app_name = 'calculations'

urlpatterns = [
    path('<str:unit_system>/formulas/validate/', ValidateViewSet.as_view()),
    path('<str:unit_system>/formulas/calculate/', CalculationView.as_view()),
]



#!/usr/bin/env python3
import os
import subprocess
from datetime import date

from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


def get_version(app):
    version = date.today().strftime('%Y-%m')
    git_tag = "0.0"
    git_commits = "0"
    suffix = "dev"
    try:
        branch = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"]
        ).rstrip().decode('utf8')
        git_describe = subprocess.check_output(
            ["git", "describe", "--long"]
        ).rstrip().decode('utf8')
        git_tag = git_describe.split('-')[0]
        git_commits = git_describe.split('-')[1]
        if branch == 'main':
            suffix = ''
        else:
            suffix = 'dev'
        print(branch, git_tag, git_commits, suffix)
        version = '{}.{}{}'.format(git_tag, git_commits, suffix)
    except (subprocess.CalledProcessError, OSError) as e:
        print('git not installed', e)
    try:
        fp = open('{}/version.py'.format(app), 'w')
        fp.write(
            'djangophysics_version = [{}, {}, "{}"]\n'.format(
                git_tag.replace('.', ', '), git_commits, suffix)
        )
        fp.close()
    except Exception:
        print('ERROR opening {}/__version__.py'.format(app), os.curdir)
    return version


module = 'djangophysics'

setup(
    name='djangophysics',
    description='Django APIs for physics conversion and calculations',
    python_requires='>3.7.0',
    version=get_version(module),
    author='Frédéric MEUROU',
    author_email='fm@peabytes.me',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://api.geophysics.io/swagger/',
    install_requires=[
        "Django~=3.0",
        "django-cors-headers~=3.2",
        "django-cors-middleware~=1.5",
        "django-createsuperuser",
        "django-extensions~=3.0",
        "django-filter~=2.3",
        "django-redis~=4.12",
        "django-sendfile~=0.3",
        "djangorestframework~=3.11",
        "drf-yasg~=1.17",
        "markdown~=3.0",
        "lxml~=4.0",
        "django_redis~=4.0",
        "pysendfile~=2.0",
        "gunicorn",
        "requests",
        # "psycopg2",
        # "mysql",
        "pytz",
        "pycountry",
        "countryinfo~=0.1",
        "timezonefinder~=4.4",
        "iso4217",
        "forex-python~=1.0",
        "Babel~=2.8",
        "Pint~=0.17",
        "networkx~=2.5",
        "sympy~=1.7",
        "channels~=3.0",
        "uncertainties~=3.1"
    ],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Environment :: Web Environment",
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Office/Business",
        "License :: OSI Approved :: MIT License",
    ],
    py_modules=[
        'djangophysics.core',
        'djangophysics.countries',
        'djangophysics.currencies',
        'djangophysics.rates',
        'djangophysics.units',
        'djangophysics.converters',
        'djangophysics.calculations'
    ],
)



"""
Serializers for Units module
"""

import logging
from datetime import date, datetime

from drf_yasg.utils import swagger_serializer_method
from rest_framework import serializers

from djangophysics.core.serializers import UserSerializer
from .models import Quantity, UnitConversionPayload, Dimension, \
    CustomUnit, Unit


class QuantitySerializer(serializers.Serializer):
    """
    Serialize a Quantity
    """
    system = serializers.CharField(label="Unit system of the Quantity")
    unit = serializers.CharField(label="Units of the quantity")
    value = serializers.FloatField(label="Magnitude of the quantity")
    date_obj = serializers.DateField(label="date of conversion")

    @staticmethod
    def validate_system(system):
        """
        Validate unit system by name
        :param system: system name
        """
        from .models import UnitSystem
        if not UnitSystem.is_valid(system):
            raise serializers.ValidationError('Invalid system')
        return system

    @staticmethod
    def validate_unit(unit):
        """
        Validating unit requires knowledge of the system
        that might embark user information for
        custom units
        Let‘s say it is valid for now
        """
        return unit

    @staticmethod
    def validate_date_obj(value):
        """
        Validate date
        :param value:
        """
        if isinstance(value, date):
            return value
        try:
            datetime.strptime(value, 'YYYY-MM-DD')
        except ValueError:
            raise serializers.ValidationError(
                'Invalid date format, use YYYY-MM-DD')
        return value

    def create(self, validated_data):
        """
        Create a Quantity object
        :param validated_data: cleaned data
        """
        return Quantity(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Quantity object
        :param instance: Quantity object
        :param validated_data: cleaned data
        """
        instance.system = validated_data.get('system', instance.system)
        instance.unit = validated_data.get('unit', instance.unit)
        instance.value = validated_data.get('value', instance.value)
        instance.date_obj = validated_data.get('date_obj', instance.date_obj)
        return instance


class UnitSerializer(serializers.Serializer):
    """
    Unit class serializer
    """
    code = serializers.CharField(label="Technical name of the unit")
    name = serializers.CharField(label="Human readable name of the unit")
    dimensions = serializers.SerializerMethodField(
        label="Dimensions the unit belongs to")

    @swagger_serializer_method(serializer_or_field=serializers.ListField)
    def get_dimensions(self, obj):
        """
        Get dimension of unit
        :param obj: Unit instance
        """
        return str(obj.dimensions)


class DimensionSerializer(serializers.Serializer):
    """
    Dimension class serializer
    """
    code = serializers.CharField(
        label="technical name of the unit (enclosed in brackets)")
    name = serializers.CharField(
        label="Human readable name of the unit")
    dimension = serializers.CharField(
        label="Mathematical expression of the dimension")
    base_unit = serializers.SerializerMethodField(
        label="Name of the base unit")

    def create(self, validated_data):
        """
        Create a Dimension object from cleaned data
        :param validated_data: cleaned data
        """
        return Dimension(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Dimension object
        :param instance: Dimension object
        :param validated_data: cleaned data
        """
        instance.unit_system = validated_data.get(
            'system',
            instance.unit_system)
        instance.code = validated_data.get('code', instance.code)
        instance.value = validated_data.get('name', instance.name)
        instance.dimension = validated_data.get('date_obj', instance.dimension)
        return instance

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_base_unit(self, obj: Dimension) -> Unit:
        """
        Get base unit for dimension object
        :param obj: Dimension object
        """
        return obj.base_unit


class DimensionWithUnitsSerializer(DimensionSerializer):
    """
    Serialize a Dimension including associated units
    """
    units = UnitSerializer(label="Units of this dimension", many=True)

    @swagger_serializer_method(serializer_or_field=UnitSerializer)
    def get_units(self, obj: Dimension) -> [Unit]:
        """
        Get units for this dimension
        """
        try:
            return obj.units()
        except KeyError as e:
            logging.error(str(e))
            return []


class UnitSystemSerializer(serializers.Serializer):
    """
    Serialize a UnitSystem
    """
    system_name = serializers.CharField()


class UnitConversionPayloadSerializer(serializers.Serializer):
    """
    Serialize a UnitConversionPayload
    """
    data = QuantitySerializer(
        label="List of unit conversions",
        many=True,
        required=False)
    base_system = serializers.CharField(
        label="Unit system used for conversion",
        required=True)
    base_unit = serializers.CharField(
        label="Unit to express the result in",
        required=True)
    batch_id = serializers.CharField(
        label="User defined batch ID",
        required=False)
    key = serializers.CharField(
        label="User defined categorization key",
        required=False)
    eob = serializers.BooleanField(
        label="End of batch ? triggers the conversion",
        default=False)
    _errors = {}

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of a converson payload
        """
        super().is_valid(raise_exception=raise_exception)
        if not self.initial_data.get('data') and \
                (not self.initial_data.get('batch_id')
                 or (self.initial_data.get('batch_id')
                     and not self.initial_data.get('eob'))):
            self._errors = {
                'data': 'data has to be provided if batch_id is not provided'
                        'or batch_id is provided and eob is False'
            }

        if self._errors and raise_exception:
            raise serializers.ValidationError(self.errors)
        return not bool(self._errors)

    @staticmethod
    def validate_base_system(value: str):
        """
        Validate unit system
        :param value: unit system name
        """
        from djangophysics.units.models import UnitSystem
        if not UnitSystem.is_valid(value):
            raise serializers.ValidationError('Invalid unit system')
        return value

    @staticmethod
    def validate_base_unit(value: str) -> str:
        """
        Validate base unit
        :param value: name of unit
        """
        from djangophysics.units.models import Unit
        if not Unit.is_valid(value):
            raise serializers.ValidationError('Invalid unit')
        return value

    def create(self, validated_data: {}) -> UnitConversionPayload:
        """
        Create a UnitConversionPayload
        :param validated_data: cleaned data
        """
        return UnitConversionPayload(**validated_data)

    def update(self, instance, validated_data: {}) -> UnitConversionPayload:
        """
        Update a UnitConversionPayload
        :param instance: UnitConversionPayload
        :param validated_data: cleaned data
        """
        self.data = validated_data.get('data', instance.data)
        self.base_system = validated_data.get(
            'base_system',
            instance.base_system)
        self.base_unit = validated_data.get('base_system', instance.base_unit)
        self.batch_id = validated_data.get('batch_id', instance.batch_id)
        self.key = validated_data.get('key', instance.key)
        self.eob = validated_data.get('eob', instance.eob)
        return instance


class CustomUnitSerializer(serializers.ModelSerializer):
    """
    Serialize a custom unit
    """
    id = serializers.CharField(
        label="ID of the CustomUnit",
        read_only=True)
    user = UserSerializer(
        label="Owner of the unit",
        read_only=True)
    unit_system = serializers.CharField(
        label="Unit system of the unit",
        read_only=True)

    class Meta:
        """
        Meta
        """
        model = CustomUnit
        fields = [
            'id',
            'user',
            'key',
            'unit_system',
            'code',
            'name',
            'relation',
            'symbol',
            'alias']



"""
Units module API filters
"""

from django.db import models
from django.db.models import QuerySet
from django_filters import rest_framework as filters

from .models import CustomUnit


class CustomUnitFilter(filters.FilterSet):
    """
    Filter on custom units
    """
    user = filters.BooleanFilter(
        label="filter rate associated to connected user",
        method='user_filter')
    key = filters.CharFilter(
        label="filter rates with key",
        method='key_filter')
    unit_system = filters.CharFilter(
        label="filter by unit system",
        field_name='unit_system',
        lookup_expr='iexact')
    code = filters.CharFilter(
        label="filter by code",
        field_name='code',
        lookup_expr='iexact')
    name = filters.CharFilter(
        label="filter by name",
        field_name='name',
        lookup_expr='icontains')
    relation = filters.CharFilter(
        label="filter by relation",
        field_name='relation',
        lookup_expr='icontains')
    symbol = filters.CharFilter(
        label="filter by symbol",
        field_name='symbol',
        lookup_expr='iexact')
    alias = models.CharField(
        "Alias",
        max_length=20,
        null=True,
        blank=True)

    ordering = filters.OrderingFilter(
        # tuple-mapping retains order
        fields=(
            ('key', 'key'),
            ('unit_system', 'unit_system'),
            ('code', 'code'),
            ('name', 'name'),
            ('relation', 'relation'),
            ('symbol', 'symbol'),
            ('alias', 'alias'),
        ),
    )

    class Meta:
        """
        Meta
        """
        model = CustomUnit
        fields = [
            'user', 'key',
            'unit_system', 'code', 'name',
            'relation', 'symbol', 'alias'

        ]

    def user_filter(self, queryset: QuerySet,
                    name: str, value: str) -> QuerySet:
        """
        Filter on request user
        """
        if self.request and self.request.user and \
                self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
            })
        return queryset.filter(user__isnull=True)

    def key_filter(self, queryset: QuerySet,
                   name: str, value: str) -> QuerySet:
        """
        Filter on key if request.user is set and authenticated
        """
        if self.request and self.request.user and \
                self.request.user.is_authenticated:
            return queryset.filter(**{
                'user': self.request.user,
                'key': value
            })
        return queryset.filter(user__isnull=True)



# Generated by Django 3.1.5 on 2021-04-19 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('units', '0004_auto_20210418_1916'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customunit',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
